//For EECE 315 Test uses
// use ./nachos -q <option> to choose test cases
// -q 1 : Default ping-pong result
// -q 2 : Test semaphore, lock and condition variable
// -q 3 : Test mailbox
// Note the Function Signature of the involved functions
// should be unified to ease the testing.

//Known Issues:
// Dynamic allocation of user-defined classes
// like Semaphore, Lock, Condition, will
// cause segementation fault.
// But declaring them as global variables
// works pretty well 

// threadtest.cc 
//	Simple test case for the threads assignment.
//
//	Create two threads, and have them context switch
//	back and forth between themselves by calling Thread::Yield, 
//	to illustratethe inner workings of the thread system.
//
// Copyright (c) 1992-1993 The Regents of the University of California.
// All rights reserved.  See copyright.h for copyright notice and limitation 
// of liability and disclaimer of warranty provisions.

#include "copyright.h"
#include "system.h"
#include "synch.h"

// testnum is set in main.cc
int testnum = 1;

Semaphore* s = new Semaphore("semaphore\0", 1);
Lock* l = new Lock("lock");
Condition* c = new Condition("condition");
Lock* cl = new Lock("condition lock");
int predicate = 0;

int sharedValue = 0;

Mailbox* mailbox = new Mailbox("mailbox");
int message_in_the_air = 0;

//----------------------------------------------------------------------
// SimpleThread
// 	Loop 5 times, yielding the CPU to another ready thread 
//	each iteration.
//
//	"which" is simply a number identifying the thread, for debugging
//	purposes.
//----------------------------------------------------------------------

void
SimpleThread(int which)
{
    int num;
    
    for (num = 0; num < 5; num++) {
	printf("*** thread %d looped %d times\n", which, num);
        currentThread->Yield();
    }
}

//----------------------------------------------------------------------
// EECE 315 Test cases
//----------------------------------------------------------------------

void SemaphoreTest(int which)
{
    int num;
    int tmp;
    for (num = 0; num < 5; num++) {
      s->P();
      tmp = sharedValue;
      currentThread->Yield();
      tmp = tmp + 1;
      currentThread->Yield();
      sharedValue = tmp;
      currentThread->Yield();
      printf("Thread %d, sharedValue %d\n", which, sharedValue);
      
      s->V();
      currentThread->Yield();
    }
}

void LockTest(int which)
{
    int num;
    int tmp;
    
    for (num = 0; num < 5; num++) {

//      l->Acquire();
      tmp = sharedValue;
      currentThread->Yield();
      tmp = tmp + 1;
      currentThread->Yield();
      sharedValue = tmp;
      currentThread->Yield();
      printf("Thread %d, sharedValue %d\n", which, sharedValue);
      
 //     l->Release();
    }
}

void CondtionTest(int which)
{
    int num;
    int tmp;
    
    for (num = 0; num < 5; num++) {

      cl->Acquire();
      while(predicate != which)
	  c->Wait(cl);
      tmp = sharedValue;
      currentThread->Yield();
      tmp = tmp + 1;
      currentThread->Yield();
      sharedValue = tmp;
      currentThread->Yield();
      printf("Thread %d, sharedValue %d\n", which, sharedValue);
      
      predicate = 1 - which;
      c->Signal(cl);
      cl->Release();
    }
}

//Test threads for mailbox
// Four threads will simulate
// sending and receiving 
// messages in an out-of-order
// way. The correct output should
// be 1 1 2 2 3 3 4 4 5 5 6 6 7 7
// as sending, receiving....order

void MailboxTest_0(int which)
{
	int message = 0;
	message_in_the_air = mailbox->Send(1);
	printf("message sent: %d\n",message_in_the_air);
	message_in_the_air = mailbox->Send(4);
	printf("message sent: %d\n",message_in_the_air);
	currentThread->Yield();
	message = mailbox->Receive(&message_in_the_air);
	printf("message received: %d\n",message);
	currentThread->Yield();
	message = mailbox->Receive(&message_in_the_air);
	printf("message received: %d\n",message);
	currentThread->Yield();
}
void MailboxTest_1(int which)
{
	int message = 0;
	message_in_the_air = mailbox->Send(5);
	printf("message sent: %d\n",message_in_the_air);
	message = mailbox->Receive(&message_in_the_air);
	printf("message received: %d\n",message);
	currentThread->Yield();
	message = mailbox->Receive(&message_in_the_air);
	printf("message received: %d\n",message);
}
void MailboxTest_2(int which)
{
	int message = 0;
	message = mailbox->Receive(&message_in_the_air);
	printf("message received: %d\n",message);
	message = mailbox->Receive(&message_in_the_air);
	printf("message received: %d\n",message);
	currentThread->Yield();
	message_in_the_air = mailbox->Send(6);
	printf("message sent: %d\n",message_in_the_air);
	currentThread->Yield();
	message_in_the_air = mailbox->Send(7);
	printf("message sent: %d\n",message_in_the_air);
}
void MailboxTest_3(int which)
{
	int message = 0;
	message_in_the_air = mailbox->Send(2);
	printf("message sent: %d\n",message_in_the_air);
	currentThread->Yield();
	message_in_the_air = mailbox->Send(3);
	printf("message sent: %d\n",message_in_the_air);
	currentThread->Yield();
	message = mailbox->Receive(&message_in_the_air);
	printf("message received: %d\n",message);
}

//----------------------------------------------------------------------
// ThreadTest1
// 	Set up a ping-pong between two threads, by forking a thread 
//	to call SimpleThread, and then calling SimpleThread ourselves.
//----------------------------------------------------------------------

void
ThreadTest1()
{
    DEBUG('t', "Entering ThreadTest1");

    Thread *t = new Thread("forked thread");

    t->Fork(SimpleThread, 1);
    SimpleThread(0);
}

//----------------------------------------------------------------------
// EECE 315 Test cases
//----------------------------------------------------------------------

void
ThreadTestSemaphore()
{
    DEBUG('t', "Entering ThreadTestSemaphore");
    //sharedValue = 0;
    Thread *t = new Thread("forked SemaphoreTest");
    
    //Semaphore* s = new Semaphore("semaphore\0", 1);
    t->Fork(SemaphoreTest, 1);
    SemaphoreTest(0);
}

void
ThreadTestLock()
{
    DEBUG('t', "Entering ThreadTestLock");

    Thread *t = new Thread("forked LockTest");

    //Lock* l = new Lock("lock");
    t->Fork(LockTest, 1);
    LockTest(0);
}

void
ThreadTestCondition()
{
    DEBUG('t', "Entering ThreadTestLock");

    Thread *t = new Thread("forked CondtionTest");

    //Condition* c = new Condition("condition");
    //Lock* cl = new Lock("condition lock");
    t->Fork(CondtionTest, 1);
    CondtionTest(0);
}

void
ThreadMailbox()
{
	DEBUG('t', "Entering ThreadMailbox");
	Thread *t1 = new Thread("forked mailbox_1");
	Thread *t2 = new Thread("forked mailbox_2");
	Thread *t3 = new Thread("forked mailbox_3");
	
	//Mailbox* mailbox = new Mailbox("mailbox");
	t1->Fork(MailboxTest_1, 1);
	t2->Fork(MailboxTest_2, 2);
	t3->Fork(MailboxTest_3, 3);
	MailboxTest_0(0);
}

void Print(int i) {
    int j = 0;
    for(j=0; j<5;j++) {
        printf("Counter value %d\n", i);
        i++;
    }
}

void
ThreadTestJoin()
{
    DEBUG('t',"Entering ThreadTestJoin");
    Thread *t1 = new Thread("forked 1 join test");
    t1->Fork(Print,1);
    
    Thread *t2 = new Thread("forked 2 join test");
    t2->Fork(Print,6);
    
    t2->Join();
    t1->Join();
    
    Print(11);
}

//----------------------------------------------------------------------
// ThreadTest
// 	Invoke a test routine.
//----------------------------------------------------------------------

void
ThreadTest()
{
    switch (testnum) {
        case 1:
            ThreadTest1();
            break;
        case 2:
            ThreadTestSemaphore();
            break;
        case 3:
            ThreadTestLock();
            break;
        case 4:
            ThreadTestCondition();
            break;
        case 5:
            ThreadMailbox();
            break;
        case 6:
            ThreadTestJoin();
            break;
        default:
            printf("No test specified.\n");
            break;
    }
}


